from laddu.amplitudes import Amplitude, ParameterLike

def Scalar(name: str, value: ParameterLike) -> Amplitude: ...
def ComplexScalar(name: str, re: ParameterLike, im: ParameterLike) -> Amplitude: ...
def PolarComplexScalar(
    name: str, r: ParameterLike, theta: ParameterLike
) -> Amplitude: ...
